/*  
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <assert.h>
#include "h/htn.h"
#include "../h/conf.h"

using namespace std;

/* Constructor */
HyperTemporalNetwork::HyperTemporalNetwork(const char* filename){
	load(filename);
	assert(this->is_well_defined());
}

/* Another Constructor */
HyperTemporalNetwork::HyperTemporalNetwork(unsigned long n){
	assert(n>0);
	this->n = n;
	init_hyperarcs();
}

/* Destructor */
HyperTemporalNetwork::~HyperTemporalNetwork(){
	if(VERBOSE_MODE) o_stream << "Destroying HTN... ";
	this->delete_hyperarcs();
	if(VERBOSE_MODE) o_stream << "done!"<< endl;
}

/* returns n*/
unsigned long HyperTemporalNetwork::get_n(){
	return this->n;
}

/* returns e*/
unsigned long HyperTemporalNetwork::get_e(){
	return this->e;
}

/* returns m*/
unsigned long HyperTemporalNetwork::get_m(){
	return this->m;
}

/* returns hyperarcs for tail @head_idx */
list<t_hyperarc*>* HyperTemporalNetwork::get_hyperarcs(unsigned long head_idx){
	return &this->hyperarcs[head_idx];
}

/**
* Initializes @this HTN by loading input file @filename
* see "dat/htn.dat" for an HTN file format example.
**/
void HyperTemporalNetwork::load(const char* filename){
	/*
	* init_steps:
	*	0: read number of nodes
	*	1: read number of hyperarcs 
	* 	2: read hyperarcs  
	*/
	int init_step = 0;
	/**/
	if(VERBOSE_MODE) o_stream << endl << "Verbose mode on"<<endl; else o_stream << endl <<"Verbose mode off"<<endl;
	if(VERBOSE_MODE) o_stream << "Loading Hyper Temporal Network from input file: " << filename << "..." << endl;
	ifstream input(filename);
	string line;
	unsigned int arc_counter = 0;
	if(input.is_open()){
		unsigned long long num_line = 0;
		long long num_nodes, num_arcs;
		while(getline(input, line)){
			num_line++;
			if(line.compare(0,1,"#")==0){ // skip comments
				if(VERBOSE_MODE) o_stream << "   ... comment found at line #" << num_line << endl;
				continue;
			}
			istringstream iss(line);
			switch(init_step){
				case 0:
					if(VERBOSE_MODE) o_stream 
						<< "   ... reading node's number at line #" << num_line;
					iss >> num_nodes;
					assert(num_nodes >= 0);
					this->n = num_nodes;
					this->init_hyperarcs();	
					init_step++;
					if(VERBOSE_MODE) o_stream << ": " << num_nodes << endl; 
				break;
				case 1:
					if(VERBOSE_MODE) o_stream 	
						<< "   ... reading hyperarc's number at line #" << num_line;
					iss >> num_arcs;
					assert(num_arcs >= 0);
					init_step++;
					if(VERBOSE_MODE) o_stream << ": " << num_arcs << endl;
				break;
				case 2:
					if(VERBOSE_MODE) o_stream 
						<< "   ... reading hyperarc data at line #" 
						<< num_line << ": "<<endl ;
					assert(arc_counter < num_arcs);
					t_hyperarc* new_hyperarc = new t_hyperarc;
					long long tail;
					iss >> tail;
					assert(tail >= 0 && tail < this->n);
					new_hyperarc->tail = tail;
					if(VERBOSE_MODE) o_stream << "\t\t\t\t head " << tail << endl;
					long long head, head_weight;
					while(iss >> head >> head_weight){
						assert(head >= 0 && head < this->n);
						assert(head != tail);
						pair<unsigned long, long> entry(head, head_weight);
						new_hyperarc->weighted_heads.insert(entry);
						if(VERBOSE_MODE) 
							o_stream << "\t\t\t\t tail " 
								<< head << " weighted " << head_weight << endl; 
					}	
					push_hyperarc(tail, new_hyperarc);
					arc_counter++;
					if(VERBOSE_MODE) o_stream << "   ... hyperarc added. " << endl;
				break;
			}
		}
		input.close();
	}else throw "cannot open input file";
	if(VERBOSE_MODE) o_stream << "... Hyper Temporal Network load completed!" << endl << endl;
}

/* initializes hyperarcs array list */
void HyperTemporalNetwork::init_hyperarcs(){
	unsigned long size = this->n;
	this->hyperarcs = new list<t_hyperarc*>[size];
	this->e = 0;
	this->m = 0;
}

/* pushes @hyperarc to @head list and updates arc's counters */
void HyperTemporalNetwork::push_hyperarc(unsigned long tail, t_hyperarc* hyperarc){
	assert(!hyperarc->weighted_heads.empty());
	hyperarc->idx = this->e++;
	this->m += hyperarc->weighted_heads.size();
	this->hyperarcs[tail].push_back(hyperarc);
}

/* deletes hyperarcs from memory */
void HyperTemporalNetwork::delete_hyperarcs(){
	for(unsigned long i=0; i < this->n; i++){
		list<t_hyperarc*>* hyperarc_list = &this->hyperarcs[i];
		while(!hyperarc_list->empty()){
			t_hyperarc* hyperarc = hyperarc_list->front();
			hyperarc_list->pop_front();
			delete hyperarc;
		}
	}
	delete [] this->hyperarcs;
}

/* checks whether @this is not an empty HTN */
bool HyperTemporalNetwork::is_well_defined(){
	if(this->n == 0) return false;
	return true;	
}

/* prints @this HTN to file output stream */
void HyperTemporalNetwork::print(){
	o_stream << "## Hyper Temporal Network definition: "<< endl;
	o_stream << this->n<< endl;
	o_stream << this->e << endl;
	for(unsigned long i=0; i < this->n; i++){
		list<t_hyperarc*>* hyperarc_list = &this->hyperarcs[i];
		list<t_hyperarc*>::iterator it = hyperarc_list->begin();
		for (it; it != hyperarc_list->end(); it++){
			t_hyperarc* hyperarc = *it;
			o_stream << hyperarc->tail << " ";
			map<unsigned long, long>::iterator it = hyperarc->weighted_heads.begin();
			for(it; it != hyperarc->weighted_heads.end(); it++)
				o_stream << it->first << " " << it->second << " ";	
			o_stream << endl;
		}
	}
	o_stream << "## End of HTN definition." << endl<<endl;
}

/* asserts whether @t is a feasible scheduling for @this */
void HyperTemporalNetwork::assert_feasible_sched(t_feasible_sched* t){
	for(unsigned long u=0; u<this->n; u++){ // for each tail u
		list<t_hyperarc*>* hyperarcs = get_hyperarcs(u);
		list<t_hyperarc*>::iterator it = hyperarcs->begin();
		for(it; it != hyperarcs->end(); it++){ // for each hyperarc with tail u
			t_hyperarc* hyperarc = *it;
			map<unsigned long, long>::iterator it = hyperarc->weighted_heads.begin();
			long min_head = t->p[it->first] - it->second; 
			for(it; it != hyperarc->weighted_heads.end(); it++){ // compute min head 
				long candidate_min_head = t->p[it->first] - it->second; 	
				if(min_head > candidate_min_head)
					min_head = candidate_min_head;
			}
			assert(t->p[u] - min_head >= 0); // and assert that reduced slack is non-negative
		}
	}	
}

/* asserts whether @neg_cyc is a negative cycle of @this */
void HyperTemporalNetwork::assert_neg_cycle(t_neg_cycle* neg_cyc){
	// assert is cycle
	map<unsigned long, t_hyperarc*>::iterator it = neg_cyc->SC.begin();
	for(it; it != neg_cyc->SC.end(); it++){
		unsigned long tail_idx = it->first;
		t_hyperarc* hyperarc = it->second;
		map<unsigned long, long>::iterator it2 = hyperarc->weighted_heads.begin();
			for(it2; it2 != hyperarc->weighted_heads.end(); it2++)
				assert(neg_cyc->SC.count(it2->first));
	}
	// assert is negative: todo
}
