/*  
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <assert.h>
#include "h/reduction.h"
#include "../h/conf.h"

using namespace std;

/* Reduces an HTN @htn to a corresponding MPG*/
MeanPayoffGame* Reduction::reduce2mpg(HyperTemporalNetwork* htn){
	if(VERBOSE_MODE) o_stream << "Reducing HTN to MPG..." ;
	unsigned long n_0 = htn->get_n(); // Min nodes
	unsigned long n_1 = htn->get_e(); // Max nodes
	MeanPayoffGame* mpg = new MeanPayoffGame(n_0, n_1);
	for(unsigned long u = 0; u < n_0; u++){ // for each tail of hyperarc
		list<t_hyperarc*>* hyperarc_list = htn->get_hyperarcs(u);
		if(hyperarc_list->empty()){ // then add a self loop arc
			t_w_arc loop_arc;
			loop_arc.head_idx = u; //loop
			loop_arc.weight = 0; // weighted zero
			mpg->push_arc(u, loop_arc);	
		}
		else{ // let assume at least one hyperarc with tail u
			list<t_hyperarc*>::iterator it = hyperarc_list->begin();
			for (it; it != hyperarc_list->end(); it++){ //iterate over hyperarcs with tail u
				t_hyperarc* hyperarc = *it;
				unsigned long hyperarc_idx = n_0 + hyperarc->idx;
				// add an arc from tail vertex "u" to "hyperarc" vertex
				t_w_arc tail_arc;
				tail_arc.head_idx = hyperarc_idx; // arc's head index 
				tail_arc.weight = 0;
				mpg->push_arc(u, tail_arc);  
				// add an arc for each head node
				map<unsigned long, long>::iterator it2 = hyperarc->weighted_heads.begin();
				for(it2; it2 != hyperarc->weighted_heads.end(); it2++){
					t_w_arc head_arc;
					head_arc.head_idx = it2->first; // arc's head index
					head_arc.weight = it2->second; // arc weight
					mpg->push_arc(hyperarc_idx, head_arc);
				}
			}
		}
	}
	assert(mpg->is_well_defined());
	if(VERBOSE_MODE) o_stream << "... done!" << endl<<endl;
	return mpg;
}

/* Reduces an MPG to a corresponding HTN*/
HyperTemporalNetwork* Reduction::reduce2htn(MeanPayoffGame* mpg){
	if(VERBOSE_MODE) o_stream << "Reducing MPG to HTN..." ;
	unsigned long n = mpg->get_n_0() + mpg->get_n_1();
	HyperTemporalNetwork* htn = new HyperTemporalNetwork(n);
	for(unsigned long u = 0; u < n; u++){
		list<t_w_arc>* arc_list = mpg->get_arcs(u);
		list<t_w_arc>::iterator it = arc_list->begin();
		if(u < mpg->get_n_0()) // black Min vertex ==> multiple small hyperarcs 
			for (it; it != arc_list->end(); it++){ // iterate over arcs with head "u"
				t_w_arc arc = *it;
				t_hyperarc* hyperarc = new t_hyperarc;
				hyperarc->tail = u;
				pair<unsigned long, long> head(arc.head_idx, arc.weight);
				hyperarc->weighted_heads.insert(head);
				htn->push_hyperarc(u, hyperarc);						
			}
		else{ // white Max vertex ==> 1! big hyperarc
			t_hyperarc* hyperarc = new t_hyperarc;
			hyperarc->tail = u;
			for(it; it != arc_list->end(); it++){ // iterate over arcs with head "u"
				t_w_arc arc = *it;
				pair<unsigned long, long> head(arc.head_idx, arc.weight);
				hyperarc->weighted_heads.insert(head);
			}		
			htn->push_hyperarc(u, hyperarc);
		}	
	}
	if(VERBOSE_MODE) o_stream << "... done!" << endl << endl;
	return htn;
}
