/*  
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "h/solver.h"
#include <climits>
#include <assert.h>
using namespace std;

/* internal solver function prototypes  */
t_feasible_sched* compute_feasible_sched(t_graph* G, unsigned long n_0);
t_feasible_sched* compute_feasible_sched_BFTest(t_graph* G, unsigned long n_0);
void insert_corresponding_hyperarc(HyperTemporalNetwork* htn, MeanPayoffGame* mpg, t_neg_cycle* neg_cycle, t_w_arc arc, unsigned long u);
MeanPayoffGame* compute_G_W_0(MeanPayoffGame* mpg, bool* decision);
unsigned long add_fake_source(t_graph* G, unsigned long n_0);
bool arc_relax(unsigned long tail, bool* infinite_distance, t_w_arc arc, t_feasible_sched* t);
void delete_graph(t_graph* G);
/**/

/* decides whether @htn is consistent or not */
bool is_consistent(HyperTemporalNetwork* htn){
	Reduction reduction;
	MeanPayoffGame *mpg = reduction.reduce2mpg(htn);
	mpg->set_algo(ALGO_ENGINE);
	bool consistent = mpg->is_all_winning(MAX);
	delete mpg;
	return consistent;
}

/* computes and returns a feasible scheduling for @htn */
/* pre-cond: assume @htn is consistent */
t_feasible_sched* compute_feasible_sched(HyperTemporalNetwork* htn){
	Reduction reduction;
	MeanPayoffGame *mpg = reduction.reduce2mpg(htn);
	mpg->set_algo(ALGO_ENGINE);
	t_feasible_sched* htn_sched; 
	t_graph* G; 
	unsigned long s_idx;
	unsigned long* energy;
	switch(ALGO_ENGINE){
	case ZP_CLASSICAL:
		G = mpg->compute_optimal_strategy_subgraph(MAX);
		s_idx = add_fake_source(G, mpg->get_n_0());
		htn_sched = compute_feasible_sched(G, mpg->get_n_0());	
		delete_graph(G);
	break;
	case ZP_IMPROVED:
		G = mpg->compute_winning_strategy_subgraph(MAX);
		s_idx = add_fake_source(G, mpg->get_n_0());
		htn_sched = compute_feasible_sched(G, mpg->get_n_0());	
		delete_graph(G);
	break;
	case BRIM2011:
		energy = mpg->solve_energy();
		htn_sched->size = mpg->get_n_0();
		htn_sched->p = new long[mpg->get_n_0()];
		for(unsigned long u=0; u < mpg->get_n_0(); u++) //schedule MIN nodes
			htn_sched->p[u] = energy[u];
		delete [] energy;					
	break;
	}
	delete mpg;
	return htn_sched;
}

t_feasible_sched* compute_feasible_sched(MeanPayoffGame* mpg, unsigned long* energy){
	t_feasible_sched* htn_sched = new t_feasible_sched;
	t_graph* G;
	unsigned long s_idx;
	switch(ALGO_ENGINE){
	case ZP_CLASSICAL:
	case ZP_IMPROVED:
		throw "ZP algorithm not compatible with energy argument";
	break;
	case BRIM2011:
		htn_sched->size = mpg->get_n_0();
		htn_sched->p = new long[mpg->get_n_0()];
		for(unsigned long u=0; u < mpg->get_n_0(); u++) //schedule MIN nodes
			htn_sched->p[u] = energy[u];
	break;
	}
	return htn_sched;
}

t_feasible_sched* compute_feasible_sched_BFTest(MeanPayoffGame* mpg){
	t_graph* G = mpg->compute_winning_strategy_subgraph(MAX);
	unsigned long s_idx = add_fake_source(G, mpg->get_n_0());
	t_feasible_sched* htn_sched = compute_feasible_sched_BFTest(G, mpg->get_n_0());
	delete_graph(G);
	return htn_sched;
}

/* computes a negative cycle of @htn */
/* pre-cond: assume htn is not consistent */
t_neg_cycle* compute_neg_cycle(HyperTemporalNetwork* htn){
	Reduction reduction;
	MeanPayoffGame *mpg = reduction.reduce2mpg(htn);
	mpg->set_algo(ALGO_ENGINE);
//	mpg->print();	
	t_neg_cycle* neg_cycle;
	bool* decision;
	decision = mpg->solve_decision();
	neg_cycle = compute_neg_cycle(htn, mpg, decision);
	delete [] decision;
	delete mpg;
	return neg_cycle;
}

t_neg_cycle* compute_neg_cycle(HyperTemporalNetwork* htn, MeanPayoffGame* mpg, bool* decision){
	MeanPayoffGame *GW_0 = compute_G_W_0(mpg, decision);	
	t_w_arc* pi = GW_0->compute_winning_strategy(MIN);
	t_neg_cycle* neg_cycle = new t_neg_cycle;
	for(unsigned long u=0; u < GW_0->get_n_0(); u++){
		t_w_arc arc = pi[u];
		insert_corresponding_hyperarc(htn, mpg, neg_cycle, arc, u);
	}	
	delete GW_0;
	delete [] pi;
	return neg_cycle;	
}

MeanPayoffGame* compute_G_W_0(MeanPayoffGame* mpg, bool* decision){
	unsigned long n_0=0, n_1=0;
	unsigned long size = mpg->get_n_0() + mpg->get_n_1();
	list<unsigned long> *W_0 = new list<unsigned long>;
	int *contains = new int[size];
	for(unsigned long u=0; u < size; u++){
		contains[u]=-1;
		if(u < mpg->get_n_0() && !decision[u]){ 
			contains[u] = n_0;	
			W_0->push_back(u);
			n_0++;
		}
		if(u >= mpg->get_n_0() && !decision[u]){ 
			contains[u] = n_0 + n_1;
			W_0->push_back(u);
			n_1++;
		}
	}
	MeanPayoffGame* GW_0 = new MeanPayoffGame(n_0, n_1);
	GW_0->set_algo(ALGO_ENGINE);
	list<unsigned long>::iterator it = W_0->begin();
	for(unsigned long u=0; it != W_0->end(); it++, u++){
		unsigned long current_node = *it;
		list<t_w_arc>* arcs = mpg->get_arcs(current_node);
		list<t_w_arc>* W_0arcs = new list<t_w_arc>;
		list<t_w_arc>::iterator it2 = arcs->begin();
		for(it2; it2 != arcs->end(); it2++){
			t_w_arc current_arc = *it2;
			unsigned long head_idx = current_arc.head_idx;
			if(contains[head_idx]>=0){
				t_w_arc new_arc;
				new_arc.head_idx = contains[head_idx];
				new_arc.weight = current_arc.weight;	
				W_0arcs->push_back(new_arc);
			}
		}
		GW_0->set_arcs(u, W_0arcs);
	}
	delete W_0; 
	delete [] contains;
	return GW_0;
}

void insert_corresponding_hyperarc(HyperTemporalNetwork* htn, MeanPayoffGame* mpg, t_neg_cycle* neg_cycle, t_w_arc arc, unsigned long u){
	unsigned long hyperarc_idx = arc.head_idx - mpg->get_n_0();
	list<t_hyperarc*>* hyper_arcs = htn->get_hyperarcs(u);
	list<t_hyperarc*>::iterator it = hyper_arcs->begin();
	for(it; it != hyper_arcs->end(); it++){
		t_hyperarc* hyperarc = *it;
		if(hyperarc->idx == hyperarc_idx){
			pair<unsigned long, t_hyperarc*> entry(u, hyperarc);
			neg_cycle->SC.insert(entry); // insert hyperarc 
			break;
		}
	}
}

/* adds a fake source vertex to V_G, connecting it to black vertices 
with new zero-weighted arcs */
unsigned long add_fake_source(t_graph* G, unsigned long n_0){
	unsigned long src_idx = G->n;
	G->n++;
	list<t_w_arc>* new_arc_array = new list<t_w_arc>[G->n];
	for(unsigned long u=0; u < G->n-1; u++){
		new_arc_array[u] = G->arcs[u]; // copy arc list
		if(u < n_0){ // add arc from fake source to u
			t_w_arc new_arc;
			new_arc.head_idx = u;
			new_arc.weight = 0;
			new_arc_array[src_idx].push_back(new_arc);
		}
	}
	delete [] G->arcs;
	G->arcs = new_arc_array;		
}

/* computes and returns a feasible scheduling for @G */
/* pre-cond: assume G is conservative with source idx G->n-1 */
/* this is a Bellman-Ford based procedure*/
t_feasible_sched* compute_feasible_sched(t_graph* G, unsigned long n_0){
	t_feasible_sched t;
	t.size = G->n;
	t.p = new long[t.size];
	bool infinite_distance[t.size];
	t.p[t.size-1] = 0;
	infinite_distance[t.size-1] = false;
	for(unsigned long u; u < t.size-1; u++)
		infinite_distance[u] = true;
	for(unsigned long phase=0; phase < G->n-1; phase++){
		for(unsigned long u=0; u < G->n; u++){
			list<t_w_arc>* arc_list = &G->arcs[u];
			list<t_w_arc>::iterator it = arc_list->begin();
			for(it; it != arc_list->end(); it++){
				t_w_arc arc = *it;
				if(arc_relax(u, infinite_distance, arc, &t)){
					t.p[arc.head_idx] = t.p[u] + arc.weight;
					infinite_distance[arc.head_idx] = false;
				}
			}
		}
	}
	// fill feasible sched
	t_feasible_sched* htn_sched = new t_feasible_sched;
	htn_sched->size = n_0;
	htn_sched->p = new long[n_0];
	for(unsigned long u=0; u < n_0; u++)
		htn_sched->p[u] = - t.p[u]; // revert sign
	delete [] t.p;
	return htn_sched;
}

t_feasible_sched* compute_feasible_sched_BFTest(t_graph* G, unsigned long n_0){
	t_feasible_sched t;
	t.size = G->n;
	t.p = new long[t.size];
	bool infinite_distance[t.size];
	t.p[t.size-1] = 0;
	infinite_distance[t.size-1] = false;
	long min_value = LONG_MAX;
	for(unsigned long u; u < t.size-1; u++)
		infinite_distance[u] = true;
	bool stable = false;
	for(unsigned long phase=0; phase < G->n-1 && !stable; phase++){
		stable = true;
		for(unsigned long u=0; u < G->n; u++){
			list<t_w_arc>* arc_list = &G->arcs[u];
			list<t_w_arc>::iterator it = arc_list->begin();
			for(it; it != arc_list->end(); it++){
				t_w_arc arc = *it;
				if(arc_relax(u, infinite_distance, arc, &t)){
					stable = false;
					t.p[arc.head_idx] = t.p[u] + arc.weight;
					if(min_value > t.p[arc.head_idx]) min_value = t.p[arc.head_idx];
					infinite_distance[arc.head_idx] = false;
				}
			}
		}
	}
	// fill feasible sched
	t_feasible_sched* htn_sched = new t_feasible_sched;
	htn_sched->size = n_0;
	htn_sched->p = new long[n_0];
	for(unsigned long u=0; u < n_0; u++)
		htn_sched->p[u] = t.p[u] - min_value; // get positive potential 
	delete [] t.p;
	return htn_sched;
}

/* relaxes @arc with respect to its @tail and distance function @t*/
bool arc_relax(unsigned long tail, bool* infinite_distance, t_w_arc arc, t_feasible_sched* t){
	if(infinite_distance[tail]) return false;
	long candidate_distance = t->p[tail] + arc.weight;
	return (infinite_distance[arc.head_idx] || 
		candidate_distance < t->p[arc.head_idx]) ? true : false;
}

/* deletes graph G from heap memory */
void delete_graph(t_graph* G){
	delete [] G->arcs;
	delete G;
}
