/*  
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/******************************************************
** Source Code - Consistency-HTN Tool   **
** Author: Carlo Comin, 03/2014		 	     **
*******************************************************/

#include <iostream>
#include <sys/time.h>
#include <sys/resource.h>
#include "../h/conf.h"
#include "../htn/h/htn.h"
#include "../mpg/h/mpg.h"
#include "../htn/h/reduction.h"
#include "../htn/h/solver.h"

using namespace std;

const char* CONSISTENT = "consistent.";
const char* INCONSISTENT = "inconsistent.";
const char* INPUT_FILE_MPG="data/mpg.dat";
const char* INPUT_FILE_HTN_CONSISTENT="data/htn_consistent.dat";
const char* INPUT_FILE_HTN_INCONSISTENT="data/htn_inconsistent.dat";
const char* OUTPUT_FILE="data/output.dat";
const bool VERBOSE_MODE=false;
AlgoEnum ALGO_ENGINE = ZP_IMPROVED;
bool CHECK_CONSISTENCY = false;
bool COMPUTE_CERTIFICATES = false;

ofstream o_stream;
/*********************************************
* Prototypes
*********************************************/
void test1();
void test2();
void test3();
void pre_args(int argc, char** argv); 
void print_feasible_sched(t_feasible_sched* t);
void print_neg_cycle(t_neg_cycle* c);
clock_t pre_test(int n);
void post_test(int n, clock_t t_start);

void welcome_msg();
/********************************************
** Consistency-HTN Tool main()
*********************************************/
int main(int argc, char** argv){
	o_stream.open(OUTPUT_FILE, ofstream::app);
	pre_args(argc, argv);
//	welcome_msg();
	test1();
	o_stream.close();
	return 0;
} 

/*******************
** default test
*******************/
void zp_htn_test(HyperTemporalNetwork* htn){
	htn->print();
	cout << "\tDeciding HTN: ... " << endl;
	bool consistent = CHECK_CONSISTENCY ? is_consistent(htn) : true;
	string str_decision = consistent ? CONSISTENT : INCONSISTENT; 
	cout << "\t\tHTN is " << str_decision << endl;
	o_stream << "\t\tHTN is " << str_decision << endl; 
	if(COMPUTE_CERTIFICATES)
		if(consistent){
		//	cout << "\tComputing a feasible scheduling... "<<endl;
			t_feasible_sched *htn_sched = compute_feasible_sched(htn);
			//htn->assert_feasible_sched(htn_sched); // check htn_sched 
		//	print_feasible_sched(htn_sched);	
			delete [] htn_sched->p;
			delete htn_sched;
		//	cout << "\tFeasible scheduling computed!"<<endl;
		} else{
		//	cout << "\tComputing a negative cycle... "<<endl;
			t_neg_cycle* htn_neg_cyc = compute_neg_cycle(htn);
			//htn->assert_neg_cycle(htn_neg_cyc); // check htn_neg_cyc
		//	print_neg_cycle(htn_neg_cyc);	
			delete htn_neg_cyc;
		//	cout << "\tNegative cycle computed!"<<endl;
		}
}

void brim_htn_test(HyperTemporalNetwork* htn){
//	htn->print();
	//cout << "\tDeciding HTN: ... " << endl;
	// get energy levels
	Reduction reduction;
	MeanPayoffGame *mpg = reduction.reduce2mpg(htn);
	mpg->set_algo(ALGO_ENGINE);
	unsigned long* energy = mpg->solve_energy();
	unsigned long size = mpg->get_n_0() + mpg->get_n_1();
	//check consistency
	bool* decision = new bool[size];
	bool consistent = true;
	for(unsigned long u=0; u < size; u++){
		if(energy[u]==ULONG_MAX){ 
			consistent=false;
			decision[u]=false;
		} else decision[u]=true;
	}//	
	string str_decision = consistent ? CONSISTENT : INCONSISTENT;
//	cout << "\t\tHTN is " << str_decision << endl;
//	o_stream << "\t\tHTN is " << str_decision << endl; 

	//if(!consistent){	
		struct rusage usage;
		getrusage(RUSAGE_SELF, &usage);
		struct timeval real_time, utime, stime;
		utime = usage.ru_utime;
		stime = usage.ru_stime;
		timeradd(&utime, &stime, &real_time);
		cout << real_time.tv_sec << "." << real_time.tv_usec << " : " << str_decision <<  endl;
	//}
	
	if(COMPUTE_CERTIFICATES)
		if(consistent){
//			cout << "\tComputing a feasible scheduling... " << endl;
			t_feasible_sched *htn_sched = compute_feasible_sched(mpg, energy);
//			htn->assert_feasible_sched(htn_sched); // check htn_sched 
//			print_feasible_sched(htn_sched);	
			delete [] htn_sched->p; delete htn_sched;
//			cout << "\tFeasible scheduling computed!" << endl;
		} else{
//			cout << "\tComputing a negative cycle... " << endl;
			t_neg_cycle* htn_neg_cyc = compute_neg_cycle(htn, mpg, decision);
//			htn->assert_neg_cycle(htn_neg_cyc); // check htn_neg_cyc
//			print_neg_cycle(htn_neg_cyc);	
			delete htn_neg_cyc;
//			cout << "\tNegative cycle computed!"<<endl;
		}
	delete decision;
	delete mpg;
}

/***********************
** test 1 mpg2htn
***********************/
void test1(){
	//clock_t t_start = pre_test(1);
	MeanPayoffGame* mpg = new MeanPayoffGame("data/pg_mpg.dat");
	Reduction reduction;
	HyperTemporalNetwork* htn = reduction.reduce2htn(mpg);
	switch(ALGO_ENGINE){
		case ZP_CLASSICAL:
		case ZP_IMPROVED:
			zp_htn_test(htn);
		break;
		case BRIM2011:
			brim_htn_test(htn);
		break;
	}
	delete mpg;
	delete htn;
//	post_test(1, t_start);
}

void test2(){
	//clock_t t_start = pre_test(2);
	MeanPayoffGame* mpg = new MeanPayoffGame("data/pg_mpg.dat");
	mpg->set_algo(ALGO_ENGINE);	
	unsigned long* potentials = mpg->solve_energy();
	//print potentials
	unsigned long size = mpg->get_n_0() + mpg->get_n_1();
	for(unsigned long u=0; u<size; u++)
		o_stream << "node: " << u << " potential: " << potentials[u] << endl;
	//free memory
	delete [] potentials;
	delete mpg;
	//post_test(2, t_start);	
}

void test3(){
	MeanPayoffGame* mpg = new MeanPayoffGame("data/mpg.dat");
//	Reduction reduction;
//	HyperTemporalNetwork* htn = reduction.reduce2htn(mpg_in);
//	delete mpg_in;
//	MeanPayoffGame* mpg = reduction.reduce2mpg(htn);
	mpg->set_algo(ALGO_ENGINE);
	unsigned long* energy = mpg->solve_energy();
	unsigned long size = mpg->get_n_0() + mpg->get_n_1();
	bool consistent = true;
	for(unsigned long u=0; u < size; u++){
		cout << "node: " << u << "potential: " << energy[u] << endl;
		if(energy[u]==ULONG_MAX){
			consistent=false;
//			break;
		}
	}
	t_graph *pi_1 = mpg->compute_winning_strategy_subgraph(MAX);
	for(unsigned int u=0; u < pi_1->n; u++){
		list<t_w_arc>* arcs = &(pi_1->arcs[u]);
		list<t_w_arc>::iterator it = arcs->begin();
		for(it; it != arcs->end(); it++){
			t_w_arc arc = *it;
			cout << "tail: " << u << " head: " << arc.head_idx << endl;
		}
	}
	string str_decision = consistent ? CONSISTENT : INCONSISTENT;
//	cout << "\t\tMPG is " << str_decision << endl;
//	o_stream << "\t\tMPg is " << str_decision << endl;
	if(COMPUTE_CERTIFICATES && consistent){
//			cout << "\tComputing a feasible scheduling... " << endl;
			clock_t t_start = pre_test(3);
			t_feasible_sched* htn_sched_BF = compute_feasible_sched_BFTest(mpg);
//			print_feasible_sched(htn_sched);
			post_test(3, t_start);
			//htn->assert_feasible_sched(htn_sched); // check htn_sched
			//htn->assert_feasible_sched(htn_sched_BF); // check htn_sched
			delete [] htn_sched_BF->p; delete htn_sched_BF;
//			cout << "\tFeasible scheduling computed!" << endl;
	}
	delete mpg;
}

clock_t pre_test(int n){
       	clock_t start = clock();
//	cout << " Executing test" << n << "...\n\tWriting output data to " 
//		<< OUTPUT_FILE << "..." << endl;
//	o_stream << " Executing test" << n << "...\n\tWriting output data to " 
//		<< OUTPUT_FILE << "..." << endl; 
	return start;
}

void post_test(int n, clock_t t_start){
//	cout << " Test" << n << "completed!" << endl;
//	o_stream << " Test" << n << "completed!" << endl;
//	clock_t t_end = clock();
//	double sec_taken = (double)(t_end - t_start)/CLOCKS_PER_SEC;
//	cout << sec_taken << endl;
//	o_stream  << sec_taken << endl;
//	cout << " Execution time: " << sec_taken << " sec" << endl << endl;
//	o_stream << " Execution time: " << sec_taken << " sec" << endl << endl;
}

/* prints to output streams the scheduling @htn_sched  */
void print_feasible_sched(t_feasible_sched* htn_sched){
	cout << "\tFeasible scheduling:" << endl;
	o_stream << "\tFeasible scheduling:" << endl; 
	for(unsigned long u=0; u < htn_sched->size; u++){
		cout << "\t\tHTN's Node #" << u 
			<< " scheduled at timepoint: " << htn_sched->p[u] << endl; 
		o_stream << "\t\tNode #" << u 
			<< " scheduled at timepoint: " << htn_sched->p[u] << endl;  
	}
}

/* prints to output streams the negative cycle @neg_cyc */
void print_neg_cycle(t_neg_cycle* neg_cyc){
	cout << "\tNegative cycle:" << endl;
	o_stream << "\tNegative cycle:" << endl;
	map<unsigned long, t_hyperarc*>::iterator it = neg_cyc->SC.begin();
	for(it; it != neg_cyc->SC.end(); it++){
		t_hyperarc* hyperarc = it->second;
		map<unsigned long, long>::iterator it2 = hyperarc->weighted_heads.begin();
		cout << "\t\tHTN's Hyperarc #" 
			<< hyperarc->idx << ": tail = " 
			<< hyperarc->tail << " head = ";
		o_stream << "\t\tHTN's Hyperarc #" 
			<< hyperarc->idx << ": tail = " 
			<< hyperarc->tail << " head = ";
		for(it2; it2 != hyperarc->weighted_heads.end(); it2++){
			cout << " " << it2->first;	
			o_stream << " " << it2->first;
		}
		cout << endl; o_stream << endl;
	}
}

void pre_args(int argc, char** argv){
	if(argc > 4) throw "arguments exception";
	for(int i=1; i < argc; i++){
		string opt = argv[i];
		if(opt=="--zp_classical") ALGO_ENGINE = ZP_CLASSICAL;
		if(opt=="--zp_improved") ALGO_ENGINE = ZP_IMPROVED;
		if(opt=="--brim2011") ALGO_ENGINE = BRIM2011;
		if(opt=="--check_consistency")
			CHECK_CONSISTENCY = true;
		if(opt=="--compute_certificates")
			COMPUTE_CERTIFICATES = true;
	}
}

/* prints to output a welcome message */
void welcome_msg(){
	cout << endl;
	cout << "*                                                     *"<<endl;
	cout << "*******************************************************"<<endl;
	cout << "  **          ---- Constraints's Code ----            **"<<endl;
	cout << "  **   Hyper Temporal Network's Consistency Tool   **"<<endl;
	cout << "*******************************************************"<<endl;
	cout << "*                                                     *"<<endl;
	cout << "*                         _____                       *"<<endl;
	cout << "*                      _.'_____`._                    *"<<endl;
	cout << "*                    . .-  12  -. .                   *"<<endl;
	cout << "*                   /,' 11      1 `.\'                 *"<<endl;                 
	cout << "*                  // 10      /   2  \\                *"<<endl;
	cout << "*                 ;;         /       ::               *"<<endl;
	cout << "*                 || 9  ----O      3 ||               *"<<endl;
	cout << "*                 ::                 ;;               *"<<endl;
	cout << "*                  \\ 8            4 //                *"<<endl;
	cout << "*                   \\`. 7       5  ,/                 *"<<endl;
	cout << "*                    '.`-.__6__.-'.'                  *"<<endl;
	cout << "*                     ((-._____.-))                   *"<<endl;
	cout << "**                    _))       ((_                  **"<<endl;
	cout << "* **                 '--'       '--'               * **"<<endl;
	cout << endl;
}
