/* 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* 
    This code is an implementation of the Value Iteration Algorithm for Mean Payoff Energy Games 
     see [Brim2011]: 
       L. Brim et. al,  Journal Formal Methods in System Design, Volume 38 Issue 2, April 2011, Pages 97-118 
                        Kluwer Academic Publishers Hingham, MA, USA 
*/

#include <iostream>
#include <climits>
#include <assert.h>
#include "math.h"
#include "h/brim.h"

using namespace std;

// constructor
Brim2011Algorithm::Brim2011Algorithm(MeanPayoffGame* mpg){
	this->mpg = mpg;
	this->M_G = this->get_M_G();
}

// destructor
Brim2011Algorithm::~Brim2011Algorithm(){}

// compute decision boolean vector
bool* Brim2011Algorithm::solve_decision(){
	unsigned long* energy = this->solve_energy();
	unsigned long size = this->mpg->get_n_0() + this->mpg->get_n_1();
	bool* decision = new bool[size];
	for(unsigned long u=0; u < size; u++){
		if(energy[u]==ULONG_MAX) decision[u]=false;
		else decision[u]=true;
	}
	return decision;
}

// computes a pre-arcs list
list<pair<unsigned long, long> >* Brim2011Algorithm::compute_pre_arcs(){
	unsigned long size = this->mpg->get_n_0() + this->mpg->get_n_1();
	list<pair<unsigned long, long> >* pre_arcs = new list<pair<unsigned long, long> >[size];
	for(unsigned long u=0; u < size; u++){
		list<t_w_arc>* arc_list = this->mpg->get_arcs(u);
		list<t_w_arc>::iterator it = arc_list->begin();
		for(it; it != arc_list->end(); it++){
			t_w_arc arc = *it;
			unsigned long head_idx = arc.head_idx;
			long weight = arc.weight;
			pair<unsigned long, long> p(u, weight);
			pre_arcs[head_idx].push_front(p);
		}
	}
	return pre_arcs;
}

unsigned long Brim2011Algorithm::get_M_G(){
	unsigned long M_G=0;
	unsigned long size = this->mpg->get_n_0() + this->mpg->get_n_1();
	for(unsigned long u=0; u < size; u++){
		list<t_w_arc>* arc_list = this->mpg->get_arcs(u);
		list<t_w_arc>::iterator it = arc_list->begin();
		long max_v = 0;
		for(it; it!=arc_list->end(); it++){
			t_w_arc arc = *it;
			long weight = arc.weight;
			if(weight<0 && -weight > max_v) max_v = -weight;
		}
		M_G += max_v;
	}
	return M_G;
}

// computes the circle-minus operator, see [Brim2011], we assumee ULONG_MAX=\top
unsigned long Brim2011Algorithm::circle_op(long a, long b){
	if(a == ULONG_MAX || (a>b && a - b > this->M_G)) return ULONG_MAX;
	return a>b ? a-b : 0;
}

// computes updated value for the count(f,v) function
// pre-condition: u is a Max node
long Brim2011Algorithm::get_count(unsigned long* energy, unsigned long u){
	if(u < this->mpg->get_n_0()) throw "u is a Min node, count() is undefined";
	unsigned long count = 0;
	list<t_w_arc>* arc_list = this->mpg->get_arcs(u);
	list<t_w_arc>::iterator it = arc_list->begin();
	for(it; it!=arc_list->end(); it++){
		t_w_arc arc = *it;
		unsigned long v = arc.head_idx;
		long weight = arc.weight;
		unsigned long rhs = circle_op(energy[v], weight);
		if(energy[u] >= rhs) count++;
	}
	return count;
}

// computes the lift operator delta(f,v), see [Brim2011]
void Brim2011Algorithm::lift_op(unsigned long* energy, unsigned long u){
	unsigned long n_0 = this->mpg->get_n_0();
	unsigned long lifted_val = u<n_0 ? 0 : ULONG_MAX;
	list<t_w_arc>* arc_list = this->mpg->get_arcs(u);
	list<t_w_arc>::iterator it = arc_list->begin();
	for(it; it!=arc_list->end(); it++){
		t_w_arc arc = *it;
		unsigned long v = arc.head_idx;
		long weight = arc.weight;
		unsigned long candidate = circle_op(energy[v], weight);
		if((u < n_0 && candidate > lifted_val) || 
			(u >= n_0 && candidate < lifted_val))
			lifted_val = candidate;
	}
	if(lifted_val > this->M_G) energy[u] = ULONG_MAX;
	else energy[u] = lifted_val;
}

// Value Iteration Algorithm for Energy Games, main loop, see [Brim2011].
unsigned long* Brim2011Algorithm::solve_energy(){
	unsigned long n_0 = this->mpg->get_n_0();
	unsigned long n_1 = this->mpg->get_n_1();
	unsigned long size = n_0 + n_1;
	unsigned long* energy = new unsigned long[size];
	long* count = new long[size];
	// compute pre arc lists
	list<pair<unsigned long, long> >* pre_arcs = compute_pre_arcs();
	// init list L
	list<unsigned long> L;
	bool contains[size];
	for(unsigned long u=0; u < size; u++){
		list<t_w_arc>* arc_list = this->mpg->get_arcs(u);
		list<t_w_arc>::iterator it = arc_list->begin();
		bool insert = u < n_0 ? false : true;
		for(it; it!=arc_list->end(); it++){
			t_w_arc arc = *it;
			if(u < n_0 && arc.weight < 0){ 
				insert = true;
				break;
			}else if(u >= n_0 && arc.weight >= 0){
				insert = false;
				break;
			}
		}
		if(insert){
			L.push_front(u); // LIFO 
			//L.push_back(u); //FIFO
			contains[u]=true;
		}else contains[u]=false;
	}
	// init energy and counter
	for(unsigned long u=0; u < size; u++) energy[u] = 0;
	for(unsigned long u=0; u < size; u++){
		count[u] = 0;
		if(u >= n_0 && !contains[u]) count[u] = get_count(energy, u);	
	}
	//iterate until L goes empty
	while(!L.empty()){
		unsigned long v = L.front(); // LIFO/FIFO
		L.pop_front(); contains[v]=false; // LIFO/FIFO
		unsigned long old = energy[v];
		lift_op(energy, v);
		if(energy[v]==ULONG_MAX) break; // STOP CRITERION at first Min Node
		if(v>=n_0) count[v] = get_count(energy, v);
		list<pair<unsigned long, long> >::iterator it = pre_arcs[v].begin();
		for(it; it!=pre_arcs[v].end(); it++){
			pair<unsigned long, long> p = *it;
			unsigned long tail = p.first;
			long weight = p.second;
			if(energy[tail] < circle_op(energy[v], weight)){
				if(tail < n_0 && contains[tail]==false){ // check Min
					L.push_front(tail); // add Min node LIFO
					//L.push_back(tail); // FIFO
					contains[tail]=true;
				}else if(tail >= n_0){ // check Max
					if(energy[tail] >= circle_op(old, weight)) 
						count[tail]--;
					if(count[tail]<=0 && contains[tail]==false){
						L.push_front(tail); // LIFO
						//L.push_back(tail); // FIFO
						contains[tail]=true;
					}
				}
			}
		}
	}
	delete [] pre_arcs;
	delete [] count;  
	return energy;
}

/*pre-cond: assume every vertex of the @mpg is winning for @player */
t_w_arc* Brim2011Algorithm::compute_winning_strategy(unsigned int player){
	t_w_arc* pi;
	MeanPayoffGame* inverted_mpg;
	switch(player){
	case MIN:
		inverted_mpg = invert_mpg();	
		pi = compute_max_winning_strategy(inverted_mpg);
		pi = invert_strategy(pi);
		delete inverted_mpg;
	break;
	case MAX:
		pi = compute_max_winning_strategy(this->mpg);	
	break;
	}		
	return pi;
}

MeanPayoffGame* Brim2011Algorithm::invert_mpg(){
	unsigned long size = this->mpg->get_n_0() + this->mpg->get_n_1();
	MeanPayoffGame* inverted_mpg = new MeanPayoffGame(this->mpg->get_n_1(), this->mpg->get_n_0());
	inverted_mpg->set_algo(BRIM2011);
	for(unsigned long u=0; u < size ; u++){
		list<t_w_arc>* arc_list;
		if(u < inverted_mpg->get_n_0()){ // MIN node in inverted mpg
			unsigned long idx = this->mpg->get_n_0() + u;
			arc_list = this->mpg->get_arcs(idx);
		}else{
			unsigned long idx = u - this->mpg->get_n_1();
			arc_list = this->mpg->get_arcs(idx);
		}
		assert(arc_list->size()>0);
		list<t_w_arc>* inverted_arc_list = invert_arc_list(arc_list);
		inverted_mpg->set_arcs(u, inverted_arc_list);
	}
	return inverted_mpg;
}

list<t_w_arc>* Brim2011Algorithm::invert_arc_list(list<t_w_arc>* arc_list){
	list<t_w_arc>* inverted_arc_list = new list<t_w_arc>;
	list<t_w_arc>::iterator it = arc_list->begin();
	for(it; it != arc_list->end(); it++){
		t_w_arc arc = *it;
		unsigned long new_head_idx;
		if(arc.head_idx < this->mpg->get_n_0()) // MIN node => MAX node
			new_head_idx = this->mpg->get_n_1() + arc.head_idx;	
		else
			new_head_idx = arc.head_idx - this->mpg->get_n_0();	
		long new_weight = - arc.weight;
		t_w_arc new_arc; 
		new_arc.head_idx = new_head_idx;
		new_arc.weight = new_weight;	
		inverted_arc_list->push_back(new_arc); 
	}
	return inverted_arc_list;
}

t_w_arc* Brim2011Algorithm::invert_strategy(t_w_arc* pi){
	unsigned long size = this->mpg->get_n_0() + this->mpg->get_n_1();
	t_w_arc* inverted_strategy = new t_w_arc[size];
	for(unsigned long u=0; u < this->mpg->get_n_0(); u++){
		t_w_arc winning_arc = pi[this->mpg->get_n_1() + u];	
		unsigned long new_head_idx;
		if(winning_arc.head_idx < this->mpg->get_n_1()) // MIN node => MAX node
			new_head_idx = this->mpg->get_n_0() + winning_arc.head_idx;	
		else
			new_head_idx = winning_arc.head_idx - this->mpg->get_n_1();
		long new_weight = - winning_arc.weight;	
		t_w_arc inverted_winning_arc;
		inverted_winning_arc.head_idx = new_head_idx;
		inverted_winning_arc.weight = new_weight;
		inverted_strategy[u] = inverted_winning_arc;	
	}
	delete [] pi;
	return inverted_strategy;
}

/* compute max winning strategy 
   pre-cond: V_G is a MAX domain
*/
t_w_arc* Brim2011Algorithm::compute_max_winning_strategy(MeanPayoffGame* G){
	unsigned long size = G->get_n_0() + G->get_n_1();
	unsigned long* energy = G->solve_energy();
	t_w_arc* pi = new t_w_arc[size];
	for(unsigned int u = G->get_n_0(); u < size; u++){
		if(energy[u]==ULONG_MAX) throw "pre-condition not met";
		list<t_w_arc> *arc_list = G->get_arcs(u);
		list<t_w_arc>::iterator it = arc_list->begin();
		for(it; it != arc_list->end(); it++){
			t_w_arc arc = *it;
			if(energy[u] >= circle_op(energy[arc.head_idx], arc.weight)){
				pi[u] = arc;
				break;
			}
		}
	} 
	return pi;
}

t_w_arc* Brim2011Algorithm::compute_optimal_strategy(){
	unsigned long size = this->mpg->get_n_0() + this->mpg->get_n_1();
	t_w_arc* p1 = new t_w_arc[size];
	throw "NOT YET IMPLEMENTED";
	//TODO
	return p1;
}
