/* 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <assert.h>
#include <algorithm>
#include "math.h"
#include "h/mpg.h"
#include "zwickpaterson/h/zwickpaterson.h"
#include "brim/h/brim.h"

using namespace std;

/*Constructor*/
MeanPayoffGame::MeanPayoffGame(const char* filename){
	load(filename);
	assert(is_well_defined());
}

/*Another Constructor*/
MeanPayoffGame::MeanPayoffGame(unsigned long n_0, unsigned long n_1){
	assert(n_0 + n_1 > 0);
	this->n_0 = n_0; // black nodes
	this->n_1 = n_1; // black nodes
	init_arcs();
}

/*Destructor*/
MeanPayoffGame::~MeanPayoffGame(){
	if(VERBOSE_MODE) o_stream << "Destroying MPG... ";
	delete_arcs();
	if(VERBOSE_MODE) o_stream << "done!"<< endl;
}

void MeanPayoffGame::set_algo(AlgoEnum algo){
	this->algo = algo;
}

AlgoEnum MeanPayoffGame::get_algo(){
	return this->algo;
}

bool* MeanPayoffGame::solve_decision(){
	bool* decision, min_found;
	ZwickPaterson1996Algorithm ZPengine(this);
	Brim2011Algorithm Bengine(this);
	switch(this->algo){
		case ZP_CLASSICAL:
		case ZP_IMPROVED:
		decision = ZPengine.solve_decision();
		break; 
		case BRIM2011:
		decision = Bengine.solve_decision();
		break;
	}
	return decision;
}

t_rational* MeanPayoffGame::solve_opt_values(){
	t_rational* opt_vals;
	switch(this->algo){
		case ZP_CLASSICAL:
		case ZP_IMPROVED:
		ZwickPaterson1996Algorithm engine(this);
		opt_vals = engine.solve_opt_values();
		break;
	}
	return opt_vals;
}

unsigned long* MeanPayoffGame::solve_energy(){
	unsigned long *progress_measure;
	Brim2011Algorithm engine(this);
	progress_measure = engine.solve_energy();	
	return progress_measure;
}

/* returns n_0 */
unsigned long MeanPayoffGame::get_n_0(){
	return this->n_0;
}

/* returns n_1 */
unsigned long MeanPayoffGame::get_n_1(){
	return this->n_1;
}

/* returns e */
unsigned long MeanPayoffGame::get_e(){
	return this->e;
}

/* computes and returns the max abs value of arc weights */
unsigned long MeanPayoffGame::get_W(){
	unsigned long W=0;
	for(unsigned long i=0; i < this->n_0 + this->n_1; i++){
		list<t_w_arc>* arc_list = get_arcs(i);
		list<t_w_arc>::iterator it = arc_list->begin();
		for(it; it != arc_list->end(); it++){
			t_w_arc arc = *it;
			long weight = arc.weight;
			if((weight > 0 && weight > W) || (weight < 0 && -weight > W))
				W = weight > 0 ? weight : -weight;
		}
	}
	return W;
}

/* returns arc list for tail index @tail_idx */
list<t_w_arc>* MeanPayoffGame::get_arcs(unsigned long tail_idx){
	return &this->arcs[tail_idx];
}

/* sets arcs @arc_list for tail @u */
void MeanPayoffGame::set_arcs(unsigned long u, list<t_w_arc>* arc_list){
	assert(arc_list->size()>0);
	this->e -= this->arcs[u].size();
	this->arcs[u] = *arc_list;
	this->e += this->arcs[u].size();
}

/**
* Initializes @this MPG by loading the input file @filename
* see "dat/mpg.dat" for an MPG file format example.
**/
void MeanPayoffGame::load(const char* filename){
	/*
	* init_steps:
	*	0: read number of vertices
	*	1: read number of arcs 
	* 	2: read arcs  
	*/
	long init_step = 0;
	/**/
	if(VERBOSE_MODE) o_stream << endl << "Verbose mode on"<<endl; else o_stream << endl <<"Verbose mode off"<<endl;
	if(VERBOSE_MODE) o_stream << "Loading Mean Payoff Game from input file: " << filename << "..." << endl;
	ifstream input(filename);
	string line;
	unsigned long arc_counter = 0;
	if(input.is_open()){
		unsigned long long num_line = 0;
		long num_white, num_black, num_arcs;
		while(getline(input, line)){
			num_line++;
			if(line.compare(0,1,"#")==0){ // skip comments
				if(VERBOSE_MODE) o_stream << "   ... comment found at line #" << num_line << endl;
				continue;
			}
			istringstream iss(line);
			switch(init_step){
				case 0:
					if(VERBOSE_MODE) o_stream 
						<< "   ... reading vertex number at line #" << num_line;
					iss >> num_black >> num_white;
					assert(num_black >= 0);
					assert(num_white >= 0);
					this->n_0 = num_black;
					this->n_1 = num_white;
					init_arcs();
					init_step++;
					if(VERBOSE_MODE) o_stream << ": black vertices " << num_black 
							<< "; white vertices " << num_white << endl;
				break;
				case 1:
					if(VERBOSE_MODE) o_stream 
						<< "   ... reading arcs number at line #" << num_line;
					iss >> num_arcs;
					assert(num_arcs >= 0);
					init_step++;
					if(VERBOSE_MODE) o_stream << ": number of arcs " << num_arcs << endl;
				break;
				case 2:
					if(VERBOSE_MODE) o_stream << "   ... reading arc data at line #" << num_line;
					assert(arc_counter < num_arcs);
					long tail, head;
					long weight;
					iss >> tail >> head >> weight; // assume indexes are absolute 
					assert(tail >= 0 && tail < this->n_0 + this->n_1);
					assert(head >= 0 && head < this->n_0 + this->n_1);
					assert(head != tail); // we assume no self loop  
					t_w_arc new_arc;
					new_arc.head_idx = head;
					new_arc.weight = weight;
					push_arc(tail, new_arc);
					arc_counter++;
					if(VERBOSE_MODE) o_stream << ": tail " << tail 
					     << " head " << head 
					     <<" weight " << weight << endl;
				break;
			}
		}
		input.close();
	}else throw "cannot open input file";
	if(VERBOSE_MODE) o_stream << "... Mean Payoff Game load completed!" << endl<<endl;
}

/* initializes arcs array list */
void MeanPayoffGame::init_arcs(){
	unsigned long size = this->n_0 + this->n_1;
	this->arcs = new list<t_w_arc>[size];
	this->e=0;
}

/* pushes #arc into tail with index tail_idx*/
void MeanPayoffGame::push_arc(unsigned long tail_idx, t_w_arc arc){
	assert(tail_idx < this->n_0 + this->n_1);
	get_arcs(tail_idx)->push_back(arc);
	this->e++;
}

/* deletes arcs from heap memory*/
void MeanPayoffGame::delete_arcs(){
	delete [] this->arcs;
}

/* checks whether @this is not empty and 
every vertex has at least one outgoing neighbour */
bool MeanPayoffGame::is_well_defined(){
	if(this->n_0 + this->n_1 == 0) return false; // do not consider empty games
	for(unsigned long i=0; i < this->n_0 + this->n_1; i++)
		if(get_arcs(i)->empty()) return false;
	return true; 	
}

/* decrements this->e */
void MeanPayoffGame::remove_arc(unsigned long u){
	get_arcs(u)->pop_front();
	this->e--;
}

/* prints to output */
void MeanPayoffGame::print(){
	o_stream << "## Mean Payoff Game definition: "<< endl;
	o_stream << this->n_0 << " " << this->n_1 << endl;
	o_stream << this->e << endl;
	for(unsigned long i=0; i < this->n_0 + this->n_1; i++){
		list<t_w_arc>* arc_list = get_arcs(i);
		list<t_w_arc>::iterator it = arc_list->begin();
		for (it; it != arc_list->end(); it++){
			t_w_arc arc = *it;
			o_stream << i << " " << arc.head_idx << " " << arc.weight << endl;
		}
	}	
	o_stream << "## End of MPG definition." << endl;
}

/* decides whether vertex u is 
   winning starting position for @player*/
bool MeanPayoffGame::is_vertex_winning(unsigned long u, unsigned int player){
	bool* decision = this->solve_decision();
	bool ret = true;
	if(player == MAX && decision[u]==false ||
	   player == MIN && decision[u]==true)
		ret=false;
	delete decision;
	return ret;
}

bool rational_equals(t_rational x, t_rational y){
	return x.num*y.denom == x.denom*y.num;
}

/* decides whether vertex @u has optimal value @value*/
bool MeanPayoffGame::is_vertex_optimal(unsigned long u, t_rational value){
	t_rational* values = this->solve_opt_values();
	t_rational opt_val = values[u];
	delete [] values;
	return rational_equals(opt_val, value);
}

/* decides whether all vertices are 
   winning starting position for @player*/
bool MeanPayoffGame::is_all_winning(unsigned int player){
	bool* decision = this->solve_decision();
	bool ret = true;
	unsigned long size = this->get_n_0() + this->get_n_1();
	for(unsigned long u=0; u < size; u++){
		if(player == MAX && decision[u]==false ||
		   player == MIN && decision[u]==true){
			ret = false; break;
		}
	}
	delete decision;
	return ret;
}

/* computes and returns a winning-strategy-induced subgraph G_{\pi} */
/* pre-cond: assume every vertex of mpg is winning for Max */
t_graph* MeanPayoffGame::compute_winning_strategy_subgraph(unsigned int player){
	// compute winning strategy for MAX
	t_w_arc* pi = this->compute_winning_strategy(player);
	t_graph* G = compute_strategy_subgraph(pi, player);
	delete [] pi;
	return G;
}

/* computes and returns a winning-strategy-induced subgraph G_{\pi} */
/* pre-cond: assumedel noce every vertex of mpg is winning for Max */
t_graph* MeanPayoffGame::compute_optimal_strategy_subgraph(unsigned int player){
	unsigned long size = this->get_n_0() + this->get_n_1();
	// compute winning/optimal strategy for MAX
	t_w_arc* pi = this->compute_optimal_strategy();
	t_graph* G = this->compute_strategy_subgraph(pi, player);	
	delete [] pi;
	return G;
}

t_graph* MeanPayoffGame::compute_strategy_subgraph(t_w_arc* pi, unsigned int player){
	unsigned long size = this->get_n_0() + this->get_n_1();
	// init graph
	t_graph* G = new t_graph;
	G->n = size;
	G->arcs = new list<t_w_arc>[size];
	// set start/end according to player
	unsigned long not_player_start, not_player_end, player_start, player_end;
	switch(player){
		case MIN:
			not_player_start = this->get_n_0();
			not_player_end = size;
			player_start = 0;
			player_end = not_player_start;
		break;
		case MAX:
			not_player_start = 0;	
			not_player_end = this->get_n_0();
			player_start = not_player_end;
			player_end = size;
		break;
	}
	// every arc out of ~player vertex
	for(unsigned long u=not_player_start; u < not_player_end; u++){
		list<t_w_arc>* arcs = this->get_arcs(u);
		G->arcs[u] = *arcs; // copy arcs references
	}
	// !1 arc out of each player vertex according to pi
	for(unsigned long u = player_start; u < player_end; u++){
		t_w_arc winning_arc = pi[u];	
		G->arcs[u].push_front(winning_arc);
	}	
	return G;
}

t_w_arc* MeanPayoffGame::compute_winning_strategy(unsigned int player){
	t_w_arc* pi; 
	ZwickPaterson1996Algorithm ZPengine(this);
	Brim2011Algorithm Bengine(this);
	switch(this->algo){
		case ZP_CLASSICAL:
		case ZP_IMPROVED:
		pi = ZPengine.compute_winning_strategy(player);
		break; 
		case BRIM2011:
		pi = Bengine.compute_winning_strategy(player);
		break;
	}
	return pi;
}

t_w_arc* MeanPayoffGame::compute_optimal_strategy(){
	t_w_arc* pi;
	ZwickPaterson1996Algorithm ZPengine(this);
	Brim2011Algorithm Bengine(this);
	switch(this->algo){
		case ZP_CLASSICAL:
		case ZP_IMPROVED:
		pi = ZPengine.compute_optimal_strategy();
		break; 
		case BRIM2011:
		pi = Bengine.compute_optimal_strategy();
		break;
	}
	return pi;
}
