/* 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include "math.h"
#include "h/zwickpaterson.h"

using namespace std;

ZwickPaterson1996Algorithm::ZwickPaterson1996Algorithm(MeanPayoffGame* mpg){
	this->mpg = mpg;
}

ZwickPaterson1996Algorithm::~ZwickPaterson1996Algorithm(){}

/* Solves Decision-MPG problem executing the procedure of Zwick and Paterson */
bool* ZwickPaterson1996Algorithm::solve_decision(){
	if(VERBOSE_MODE) o_stream << "Solving Decision-MPG by Zwick & Paterson procedure ..." << endl;
	unsigned long size = this->mpg->get_n_0() + this->mpg->get_n_1(); // wlog > 0
	unsigned long W = this->mpg->get_W();
	unsigned long long K = W > 0 ? 4*size*size*W : 1;
	long long* nu_k = this->compute_nu_k(size, K);
	bool* decision = new bool[size];
	for(unsigned long u=0; u < size; u++){ // for each vertex u, test threshold
		long long threshold = 2*size*nu_k[u] + K; 
		if(threshold > 0) 		// player Max wins
			decision[u] = true;
		else 				// player Min wins
			decision[u] = false;
		const char* winner = decision[u] ? "Max" : "Min";
		if(VERBOSE_MODE) o_stream << "\t node #" 
			<< u << " is winning for player " << winner << endl; 
	}
	delete [] nu_k;
	if(VERBOSE_MODE) o_stream << "... done!" << endl << endl;
	return decision;
}

/* Solves Decision-MPG problem executing the classical procedure of Z.P.*/
t_rational* ZwickPaterson1996Algorithm::solve_opt_values(){
	if(VERBOSE_MODE) o_stream 
		<< "Solving Values-MPG by classic Zwick & Paterson optimal-values procedure ... " << endl;
	unsigned long size = this->mpg->get_n_0() + this->mpg->get_n_1(); // wlog > 0
	unsigned long W = this->mpg->get_W();
	unsigned long K = W > 0 ? 4*size*size*size*W : 1; // one "size" factor more
	long long* nu_k = compute_nu_k(size, K); // as in ZP_solve_decision
	t_rational* values = new t_rational[size];
	for(unsigned long u=0; u < size; u++){ // for each vertex u, compute optimal value
		o_stream << nu_k[u] << " ";
		t_rational opt_value = search_for_rational(nu_k[u], K);
		values[u] = opt_value;
	}
	delete [] nu_k;
	if(VERBOSE_MODE) o_stream << "... done!" << endl << endl;
	return values;	
}

/* Searches in time O(size)=O(n) for a rational number inside (L,R) interval given by @nu_k */
t_rational ZwickPaterson1996Algorithm::search_for_rational(long long nu_k, unsigned long K){
	unsigned long size = this->mpg->get_n_0() + this->mpg->get_n_1(); // wlog > 0
	t_rational ret_q;
	for(unsigned long denom=1; denom <= size; denom++){
		t_rational R; R.num = denom*(2*nu_k*size*(size-1)+K); R.denom = 2*size*(size-1)*K;
		t_rational L; L.num = denom*(2*nu_k*size*(size-1)-K); L.denom = 2*size*(size-1)*K;
		long long floor_R = floor(1.0*R.num/R.denom);
		long long ceil_L = ceil(1.0*L.num/L.denom);
		if(floor_R == ceil_L){
			ret_q.num = floor_R;
			ret_q.denom = denom;
			return ret_q;
		}
	}	
	throw "rational value not found, something goes wrong";
}

// Computes values of the truncated games nu_k, the core of Zwick Paterson algorithm
long long* ZwickPaterson1996Algorithm::compute_nu_k(unsigned long size, unsigned long long K){
	unsigned long W = this->mpg->get_W();
	if(VERBOSE_MODE) o_stream << "W = " << W << endl;
	if(VERBOSE_MODE) o_stream << "K = " << K << endl;
	long long* nu_k = new long long[size]; 
	for(unsigned long u=0; u < size; u++)
		nu_k[u] = 0;
	for(unsigned long long k=0; k < K; k++){
		long long *next_nu_k = new long long[size];
		for(unsigned long u=0; u < size; u++){
			list<t_w_arc>* arc_list = this->mpg->get_arcs(u);
			//init value
			t_w_arc front_w_arc = arc_list->front();
			long long value = nu_k[front_w_arc.head_idx] + front_w_arc.weight;
			//compute value
			list<t_w_arc>::iterator it = arc_list->begin();
			for(it; it != arc_list->end(); it++){
				t_w_arc arc = *it;
				unsigned long v = arc.head_idx;
				long w = arc.weight;
				long long candidate_value = nu_k[v] + w; 
				if((u < this->mpg->get_n_0() && candidate_value < value) || // for Player Min 
				   (u >= this->mpg->get_n_0() && candidate_value > value))  // for Player Max
					value = candidate_value;		
			}
			next_nu_k[u] = value;
		}
		long long *to_del = nu_k;
		nu_k = next_nu_k;
		delete [] to_del;
	}
	return nu_k;
}

/*pre-cond: assume every node of the mpg is winning for @player */
t_w_arc* ZwickPaterson1996Algorithm::compute_winning_strategy(unsigned int player){
	unsigned long size = this->mpg->get_n_0() + this->mpg->get_n_1();
	unsigned long player_start, player_end;
	switch(player){
	case MIN:
		player_start = 0;
		player_end = this->mpg->get_n_0();
	break;
	case MAX:
		player_start = this->mpg->get_n_0();
		player_end = size;
	break;
	}
	t_w_arc* pi = new t_w_arc[size];
	for(unsigned long u = player_start; u < player_end; u++){
		t_w_arc winning_arc = this->binary_search_winning_arc(u, player);
		pi[u] = winning_arc;
	}
	return pi;
}

t_w_arc* ZwickPaterson1996Algorithm::compute_optimal_strategy(){
	unsigned long size = this->mpg->get_n_0() + this->mpg->get_n_1();
	t_w_arc* pi1 = new t_w_arc[size];
	t_rational* values = this->solve_opt_values();
	for(unsigned long u=0; u < size; u++){ // optimal strategy for every node
		t_rational value = values[u];
		t_w_arc	winning_arc = this->binary_search_optimal_arc(u, value);
		pi1[u] = winning_arc;
	}
	delete [] values;
	return pi1;
}

/* returns a winning arc out of @u for @player */
/* pre-cond: assume u is winning for player "int player" */
t_w_arc ZwickPaterson1996Algorithm::binary_search_winning_arc(unsigned long u, int player){
	list<t_w_arc>* old_arc_list = this->mpg->get_arcs(u);
	if(old_arc_list->size() == 1) return old_arc_list->front();
	list<t_w_arc>* first_half = new list<t_w_arc>;
	list<t_w_arc>* second_half = new list<t_w_arc>;
	list<t_w_arc>::iterator it = old_arc_list->begin();
	unsigned long i=0; // fill first and second half
	for(it; it != old_arc_list->end(); it++, i++){
		t_w_arc arc = *it;
		if(i < old_arc_list->size()/2)
			first_half->push_back(arc);
		else
			second_half->push_back(arc);
	}
	t_w_arc arc;
	this->mpg->set_arcs(u, first_half);
	if(this->mpg->is_vertex_winning(u, player)) // go recursively of first half
		arc = this->binary_search_winning_arc(u, player);
	else{ // otherwise, go recursively on second half
		this->mpg->set_arcs(u, second_half);
		arc = this->binary_search_winning_arc(u, player);
	}
	this->mpg->set_arcs(u, old_arc_list); // restore old arcs 
	delete first_half; delete second_half;
	return arc;
}

t_w_arc ZwickPaterson1996Algorithm::binary_search_optimal_arc(unsigned long u, t_rational value){
	list<t_w_arc>* old_arc_list = this->mpg->get_arcs(u);
	if(old_arc_list->size() == 1) return old_arc_list->front();
	list<t_w_arc>* first_half = new list<t_w_arc>; 
	list<t_w_arc>* second_half = new list<t_w_arc>;
	list<t_w_arc>::iterator it = old_arc_list->begin();
	unsigned long i=0; // fill first and second half
	for(it; it != old_arc_list->end(); it++, i++){
		t_w_arc arc = *it;
		if(i < old_arc_list->size()/2)
			first_half->push_back(arc);
		else
			second_half->push_back(arc);
	}
	t_w_arc arc;
	this->mpg->set_arcs(u, first_half);
	if(this->mpg->is_vertex_optimal(u, value)) // go recursively of first half
		arc = this->binary_search_optimal_arc(u, value);
	else{ // otherwise, go recursively on second half
		this->mpg->set_arcs(u, second_half);
		arc = this->binary_search_optimal_arc(u, value);
	}
	this->mpg->set_arcs(u, old_arc_list); // restore old arcs
	delete first_half; delete second_half; 
	return arc;
}

